package cn.com.qmth.stmms.api.controller.admin;

import java.text.DecimalFormat;
import java.util.HashMap;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Set;

import javax.servlet.http.HttpServletResponse;

import org.apache.commons.lang3.StringUtils;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Controller;
import org.springframework.web.bind.annotation.RequestBody;
import org.springframework.web.bind.annotation.RequestMapping;
import org.springframework.web.bind.annotation.RequestMethod;
import org.springframework.web.bind.annotation.ResponseBody;

import com.qmth.boot.core.collection.PageResult;
import com.qmth.boot.core.exception.StatusException;

import cn.com.qmth.stmms.admin.dto.MarkInfoDTO;
import cn.com.qmth.stmms.admin.dto.MarkerInfoDTO;
import cn.com.qmth.stmms.admin.vo.SubjectLibraryVO;
import cn.com.qmth.stmms.api.controller.BaseApiController;
import cn.com.qmth.stmms.biz.exam.bean.MarkInfoVo;
import cn.com.qmth.stmms.biz.exam.bean.ResultMessage;
import cn.com.qmth.stmms.biz.exam.dao.SelectiveStudentDao;
import cn.com.qmth.stmms.biz.exam.model.ExamSubject;
import cn.com.qmth.stmms.biz.exam.model.MarkGroup;
import cn.com.qmth.stmms.biz.exam.model.Marker;
import cn.com.qmth.stmms.biz.exam.service.ExamQuestionService;
import cn.com.qmth.stmms.biz.exam.service.ExamStudentService;
import cn.com.qmth.stmms.biz.exam.service.ExamSubjectService;
import cn.com.qmth.stmms.biz.exam.service.MarkGroupService;
import cn.com.qmth.stmms.biz.exam.service.MarkerService;
import cn.com.qmth.stmms.biz.exam.service.query.ExamSubjectSearchQuery;
import cn.com.qmth.stmms.biz.lock.LockService;
import cn.com.qmth.stmms.biz.mark.query.MarkLibrarySearchQuery;
import cn.com.qmth.stmms.biz.mark.service.MarkLibraryService;
import cn.com.qmth.stmms.biz.user.service.UserService;
import cn.com.qmth.stmms.biz.utils.PageUtil;
import cn.com.qmth.stmms.common.annotation.Logging;
import cn.com.qmth.stmms.common.domain.ApiUser;
import cn.com.qmth.stmms.common.enums.LibraryStatus;
import cn.com.qmth.stmms.common.enums.LockType;
import cn.com.qmth.stmms.common.enums.LogType;
import cn.com.qmth.stmms.common.enums.MarkStatus;
import cn.com.qmth.stmms.common.utils.ExportExcel;
import io.swagger.annotations.Api;
import io.swagger.annotations.ApiOperation;

@Api(tags = "评卷管理")
@Controller("adminMarkInfoController")
@RequestMapping("/api/admin/mark/info")
public class MarkInfoController extends BaseApiController {

    protected static Logger log = LoggerFactory.getLogger(MarkInfoController.class);

    public static final String SPLIT = ",";

    @Autowired
    private ExamSubjectService subjectService;

    @Autowired
    private MarkerService markerService;

    @Autowired
    private MarkGroupService groupService;

    @Autowired
    private ExamQuestionService questionService;

    @Autowired
    private UserService userService;

    @Autowired
    private MarkLibraryService libraryService;

    @Autowired
    private LockService lockService;

    @Autowired
    private ExamStudentService studentService;

    @Autowired
    private SelectiveStudentDao selectiveStudentDao;

    @ApiOperation(value = "评卷进度列表")
    @RequestMapping(value = "query", method = RequestMethod.POST)
    @ResponseBody
    public PageResult<SubjectLibraryVO> query(ExamSubjectSearchQuery query) {
        ApiUser user = getApiUser();
        int examId = getSessionExamId();
        query.setExamId(examId);
        Set<String> unFinishSet = libraryService.findSubjectUnFinishByExamId(examId);
        unFinishSet.addAll(questionService.FindSubjectCodeByExamIdAndObjectiveAndGroupNumberIsNull(examId, false));
        if (query.getFinished() != null) {
            String subjectCodeIn = org.apache.commons.lang.StringUtils.join(unFinishSet, ",");
            if (query.getFinished()) {
                query.setCodeNotIn(subjectCodeIn);
            } else {
                query.setCodeIn(subjectCodeIn);
            }
        }
        if (user.isSubjectHeader()) {
            String subjectCodeIn = org.apache.commons.lang.StringUtils.join(user.getSubjectCodeSet(), ",");
            query.setCodeIn(subjectCodeIn);
        }
        query = subjectService.findByQuery(query);

        List<SubjectLibraryVO> list = new LinkedList<SubjectLibraryVO>();
        for (ExamSubject subject : query.getResult()) {
            SubjectLibraryVO vo = new SubjectLibraryVO();
            vo.setSubject(subject);
            vo.setStudentCount(subject.getUploadCount());
            vo.setGroupCount(groupService.countByExamAndSubject(examId, subject.getCode()));
            MarkLibrarySearchQuery mQuery = new MarkLibrarySearchQuery();
            mQuery.setExamId(examId);
            mQuery.setSubjectCode(subject.getCode());
            long libraryCount = libraryService.countByQuery(mQuery);
            mQuery.addStatus(LibraryStatus.MARKED);
            long markedCount = libraryService.countByQuery(mQuery);
            vo.setMarkedCount(markedCount);
            mQuery.addStatus(LibraryStatus.ARBITRATED);
            mQuery.addStatus(LibraryStatus.INSPECTED);
            long totalMarkedCount = libraryService.countByQuery(mQuery);
            String percent = libraryCount > 0
                    ? (new DecimalFormat("####.###").format(totalMarkedCount * 100.0 / libraryCount) + "%")
                    : "0%";
            vo.setPercent(percent);
            long count = questionService.countByExamIdAndSubjectAndObjectiveAndGroupNumberIsNull(examId,
                    subject.getCode(), false);
            vo.setGroupFinish(count == 0);
            vo.setSelectiveError(selectiveStudentDao.countByExamIdAndSubjectCodeAndLessSelectiveOrNotSelective(
                    subject.getExamId(), subject.getCode(), true, true) != 0);
            list.add(vo);
        }
        return PageUtil.of(list, query);
    }

    @ApiOperation(value = "评卷进度统计")
    @RequestMapping(value = "statistic", method = RequestMethod.POST)
    @ResponseBody
    public MarkInfoVo statistic() {
        int examId = getSessionExamId();
        Set<String> unFinishSet = libraryService.findSubjectUnFinishByExamId(examId);
        MarkLibrarySearchQuery mQuery = new MarkLibrarySearchQuery();
        mQuery.setExamId(examId);
        long total = libraryService.countByQuery(mQuery);
        mQuery.addStatus(LibraryStatus.MARKED);
        mQuery.addStatus(LibraryStatus.ARBITRATED);
        mQuery.addStatus(LibraryStatus.INSPECTED);
        long markedCount = libraryService.countByQuery(mQuery);
        long subjectCount = subjectService.count(examId);
        MarkInfoVo vo = new MarkInfoVo();
        vo.setMarkedCount(markedCount);
        vo.setUnMarkedCount(total - markedCount);
        vo.setUnSubjectFinishCount(unFinishSet.size());
        vo.setSubjectFinishCount(subjectCount - unFinishSet.size());
        return vo;
    }

    @ApiOperation(value = "整体评卷进度导出")
    @Logging(menu = "整体评卷进度导出", type = LogType.EXPORT)
    @RequestMapping(value = "/export/progress", method = RequestMethod.POST)
    public void exportProgress(HttpServletResponse response) {
        ApiUser user = getApiUser();
        int examId = getSessionExamId();
        List<MarkInfoDTO> result = new LinkedList<MarkInfoDTO>();
        List<ExamSubject> list = subjectService.list(examId);
        if (user.isSubjectHeader()) {
            list = getExamSubject(examId, user);
        }
        for (ExamSubject subject : list) {
            MarkInfoDTO dto = new MarkInfoDTO(subject);
            MarkLibrarySearchQuery mQuery = new MarkLibrarySearchQuery();
            mQuery.setExamId(examId);
            mQuery.setSubjectCode(subject.getCode());
            long libraryCount = libraryService.countByQuery(mQuery);
            mQuery.addStatus(LibraryStatus.MARKED);
            long inspectedCount = libraryService.countByQuery(mQuery);
            dto.setInspectedCount(inspectedCount);
            mQuery.addStatus(LibraryStatus.ARBITRATED);
            mQuery.addStatus(LibraryStatus.INSPECTED);
            long totalMarkedCount = libraryService.countByQuery(mQuery);
            String percent = libraryCount > 0
                    ? (new DecimalFormat("####.###").format(totalMarkedCount * 100.0 / libraryCount) + "%")
                    : "0%";
            dto.setPercent(percent);
            dto.setGroupCount(groupService.countByExamAndSubject(examId, subject.getCode()));
            result.add(dto);
        }

        try {
            String fileName = "整体评卷进度.xlsx";
            new ExportExcel("整体评卷进度", MarkInfoDTO.class).setDataList(result).write(response, fileName).dispose();
        } catch (Exception e) {
            log.error(e.getMessage());
            e.printStackTrace();
            throw new StatusException("导出整体评卷进度失败!失败信息：" + e.getMessage());
        }
    }

    @ApiOperation(value = "评卷员工作量导出")
    @Logging(menu = "评卷员工作量导出", type = LogType.EXPORT)
    @RequestMapping(value = "/export/marker", method = RequestMethod.POST)
    public void exportMarker(HttpServletResponse response) {
        ApiUser user = getApiUser();
        int examId = getSessionExamId();
        List<MarkerInfoDTO> result = new LinkedList<MarkerInfoDTO>();
        List<Marker> list = markerService.getMarkCount(examId);
        if (user.isSubjectHeader()) {
            list = markerService.getMarkCount(examId, user.getSubjectCodeSet());
        }
        Map<String, ExamSubject> subjectMap = new HashMap<String, ExamSubject>();
        Map<String, String> collegeMap = new HashMap<String, String>();
        for (Marker marker : list) {
            ExamSubject subject = subjectMap.get(marker.getSubjectCode());
            if (subject == null) {
                subject = subjectService.find(marker.getExamId(), marker.getSubjectCode());
                subjectMap.put(marker.getSubjectCode(), subject);
                collegeMap.put(marker.getSubjectCode(), StringUtils
                        .join(studentService.findDistinctCollegeBySubjectCode(examId, marker.getSubjectCode()), SPLIT));
            }
            MarkGroup group = groupService.findOne(examId, marker.getSubjectCode(), marker.getGroupNumber());
            group.setQuestionList(questionService.findByExamAndSubjectAndObjectiveAndGroupNumber(examId,
                    group.getSubjectCode(), false, group.getNumber()));
            marker.setUser(userService.findById(marker.getUserId()));
            MarkerInfoDTO dto = new MarkerInfoDTO(marker, subject, group);
            dto.setTotalScore(questionService.sumTotalScoreByGroupNumber(examId, marker.getSubjectCode(), false,
                    marker.getGroupNumber()));
            dto.setCollege(collegeMap.get(marker.getSubjectCode()));
            result.add(dto);
        }

        try {
            String fileName = "评卷员工作量.xlsx";
            new ExportExcel("评卷员工作量", MarkerInfoDTO.class).setDataList(result).write(response, fileName).dispose();
        } catch (Exception e) {
            log.error(e.getMessage());
            e.printStackTrace();
            throw new StatusException("导出评卷员工作量失败!失败信息：" + e.getMessage());
        }
    }

    @ApiOperation(value = "大题关闭")
    @Logging(menu = "大题关闭", type = LogType.UPDATE)
    @RequestMapping(value = "/finish", method = RequestMethod.POST)
    @ResponseBody
    public ResultMessage finish(@RequestBody String[] codes) {
        int examId = getSessionExamId();
        String messages = "";
        for (String code : codes) {
            MarkLibrarySearchQuery mQuery = new MarkLibrarySearchQuery();
            mQuery.setExamId(examId);
            mQuery.setSubjectCode(code);
            long libraryCount = libraryService.countByQuery(mQuery);
            mQuery.addStatus(LibraryStatus.MARKED);
            mQuery.addStatus(LibraryStatus.ARBITRATED);
            mQuery.addStatus(LibraryStatus.INSPECTED);
            long markedCount = libraryService.countByQuery(mQuery);
            if (markedCount != libraryCount) {
                continue;
            }
            List<MarkGroup> groups = groupService.findByExamAndSubject(examId, code);
            for (MarkGroup group : groups) {
                if (group == null || group.getStatus() == MarkStatus.FINISH) {
                    continue;
                }
                try {
                    lockService.waitlock(LockType.GROUP, group.getExamId(), group.getSubjectCode(), group.getNumber());
                    if (group.getStatus() == MarkStatus.FORMAL && group.getLeftCount() == 0
                            && selectiveStudentDao.countByExamIdAndSubjectCodeAndLessSelectiveOrNotSelective(
                                    group.getExamId(), group.getSubjectCode(), true, true) == 0) {
                        groupService.updateStatus(examId, code, group.getNumber(), MarkStatus.FINISH,
                                group.getStatus());
                    } else {
                        messages = " " + code;
                    }
                } finally {
                    lockService.unlock(LockType.GROUP, group.getExamId(), group.getSubjectCode(), group.getNumber());
                }
            }
        }
        if (StringUtils.isNotBlank(messages)) {
            throw new StatusException(messages + " 评卷未完成或选做题异常，无法关闭分组");
        } else {
            return resultOk();
        }
    }
}
